/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// TUT4E.CPP - VGA Trainer Program 4 (in Turbo C++ 3.0)                 //
//                                                                         //
// Author        : Grant Smith (Denthor) - denthor@beastie.cs.und.ac.za    //
// Translator    : Christopher G. Mann   - r3cgm@dax.cc.uakron.edu         //
//                                                                         //
// Last Modified : December 23, 1994                                       //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

//               //
// INCLUDE FILES //
//               //

  #include <conio.h>		   // clrscr(), getch(), kbhit()
  #include <dos.h>		   // MK_FP, geninterrupt()
  #include <iostream.h>		   // _fmemset(), cout, memset(), _fmemcpy()
  #include <stdlib.h>		   // calloc(), exit(), free()

//                     //
// FUNCTION PROTOTYPES //
//                     //

  // MODE SETTING FUNCTIONS
  void SetMCGA();
  void SetText();

  // VIRTUAL SCREEN FUNCTIONS
  void SetUpVirtual();
  void ShutDown();
  void Flip();

  // UTILITY FUNCTIONS
  void Cls(unsigned char Col, unsigned char *Where);
  void Putpixel (int x, int y, unsigned char Col, unsigned int Where);
  void WaitRetrace();

  // MID-LEVEL FUNCTIONS
  void PatternDraw1();

//                              //
// GLOBAL VARIABLE DECLARATIONS //
//                              //

  // declare a pointer to the offset of the Virtual Screen
  unsigned char *vaddr = NULL;

  // declare a pointer to the offset of the VGA memory
  unsigned char *vga = (unsigned char *) MK_FP(0xA000, 0);

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                                MAIN FUNCTION                              //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

void main() {

  clrscr();

  SetMCGA();
  SetUpVirtual();
  Cls(0,vaddr);    // After you have got the memory for the virtual screen,
		   // it is usually filled with random garbage. It is always
		   // wise to clear the virtual screen directly afterwards
  PatternDraw1();
  SetText();
  ShutDown();
  clrscr();
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// SetMCGA() - This function gets you into 320x200x256 mode.               //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void SetMCGA() {
  _AX = 0x0013;
  geninterrupt (0x10);
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// SetText() - This function gets you into text mode.                      //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void SetText() {
  _AX = 0x0003;
  geninterrupt (0x10);
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Cls() - This clears the screen to the specified color, on the VGA or on //
//         the Virtual screen.                                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void Cls(unsigned char Col, unsigned char *Where) {
  _fmemset(Where, Col, 64000);
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// WaitRetrace() - This waits until you are in a Verticle Retrace.         //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void WaitRetrace() {

  _DX = 0x03DA;

  l1: asm {
	in  al,dx;
	and al,0x08;
	jnz l1;
      }

  l2: asm {
	in  al,dx;
	and al,0x08;
	jz  l2;
      }
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// SetUpVirtual() - This sets up the memory needed for the virtual screen. //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void SetUpVirtual() {

  vaddr = (unsigned char *) calloc(64000,1);
  if (vaddr == NULL) {
    SetText();
    cout << "Not enough memory available, exiting program...";
    exit(1);
  }
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// ShutDown() - This frees the memory used by the virtual screen.          //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void ShutDown() {
  free(vaddr);
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Putpixel() - This puts a pixel at X,Y using color Col, on VGA or the    //
//              Virtual Screen;                                            //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void Putpixel (int x, int y, unsigned char Col, unsigned char *Where) {
  memset(Where+(x+(y*320)),Col,1);
}


/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Flip() - This flips the virtual screen to the VGA screen.               //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void Flip() {
  _fmemcpy(vga,vaddr,64000);
}



/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// PatternDraw1() - This tests the speed of flipping by drawing two         //
//                 patterns and flipping them.                             //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

void PatternDraw1() {

  int loop1, loop2;

  // This draws pattern to the VGA screen

  for(loop1=1; loop1<256; loop1++)
    for(loop2=1; loop2<199; loop2++)
	Putpixel (loop1,loop2,loop1,vga);
  getch();
}

